import logging
from typing import Callable, Optional

from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable

from ..common import is_script_tag, is_sequence_header
from ..models import FlvHeader, FlvTag
from .typing import FLVStream, FLVStreamItem

__all__ = ('correct',)

logger = logging.getLogger(__name__)


def correct() -> Callable[[FLVStream], FLVStream]:
    def _correct(source: FLVStream) -> FLVStream:
        """Correct the timestamp offset of the FLV tags."""

        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            disposed = False
            subscription = SerialDisposable()

            delta: Optional[int] = None
            first_data_tag: Optional[FlvTag] = None

            def reset() -> None:
                nonlocal delta, first_data_tag
                delta = None
                first_data_tag = None

            def correct_ts(tag: FlvTag, delta: int) -> FlvTag:
                if delta == 0:
                    return tag
                return tag.evolve(timestamp=tag.timestamp + delta)

            def on_next(item: FLVStreamItem) -> None:
                nonlocal delta
                nonlocal first_data_tag

                if isinstance(item, FlvHeader):
                    delta = None
                    first_data_tag = None
                    observer.on_next(item)
                    return

                tag = item

                if is_script_tag(tag):
                    tag = correct_ts(tag, -tag.timestamp)
                    observer.on_next(tag)
                    return

                if delta is None:
                    if is_sequence_header(tag):
                        tag = correct_ts(tag, -tag.timestamp)
                        observer.on_next(tag)
                    else:
                        if first_data_tag is None:
                            first_data_tag = tag
                            logger.debug(f'The first data tag: {first_data_tag}')
                        else:
                            second_data_tag = tag
                            logger.debug(f'The second data tag: {second_data_tag}')
                            if second_data_tag.timestamp >= first_data_tag.timestamp:
                                delta = -first_data_tag.timestamp
                                logger.debug(f'Timestamp delta: {delta}')
                                observer.on_next(correct_ts(first_data_tag, delta))
                                observer.on_next(correct_ts(second_data_tag, delta))
                            else:
                                delta = -second_data_tag.timestamp
                                logger.debug(f'Timestamp delta: {delta}')
                                observer.on_next(correct_ts(second_data_tag, delta))
                                observer.on_next(correct_ts(first_data_tag, delta))
                            first_data_tag = None
                    return

                tag = correct_ts(tag, delta)
                observer.on_next(tag)

            def dispose() -> None:
                nonlocal disposed
                disposed = True
                reset()

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)

    return _correct



from __future__ import annotations

import logging
import os
from typing import Optional

from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable
from tqdm import tqdm

from blrec.bili.live import Live
from blrec.flv.operators.typing import FLVStream, FLVStreamItem

__all__ = ('ProgressBar',)


DISPLAY_PROGRESS = bool(os.environ.get('BLREC_PROGRESS'))

logger = logging.getLogger(__name__)


class ProgressBar:
    def __init__(self, live: Live) -> None:
        self._live = live
        self._pbar: Optional[tqdm] = None

    def update_bar_info(self) -> None:
        if self._pbar is not None:
            self._pbar.set_postfix_str(self._make_pbar_postfix())

    def __call__(self, source: FLVStream) -> FLVStream:
        return self._progress(source)

    def _progress(self, source: FLVStream) -> FLVStream:
        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            subscription = SerialDisposable()

            self._pbar = tqdm(
                desc='Recording',
                unit='B',
                unit_scale=True,
                unit_divisor=1024,
                postfix=self._make_pbar_postfix(),
                disable=not DISPLAY_PROGRESS,
            )

            def on_next(item: FLVStreamItem) -> None:
                if self._pbar is not None:
                    self._pbar.update(len(item))
                observer.on_next(item)

            def on_completed() -> None:
                if self._pbar is not None:
                    self._pbar.close()
                    self._pbar = None
                observer.on_completed()

            def dispose() -> None:
                if self._pbar is not None:
                    self._pbar.close()
                    self._pbar = None

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)

    def _make_pbar_postfix(self) -> str:
        return '{room_id} - {user_name}: {room_title}'.format(
            room_id=self._live.room_info.room_id,
            user_name=self._live.user_info.name,
            room_title=self._live.room_info.title,
        )



class FlvDataError(ValueError):
    ...


class FlvHeaderError(FlvDataError):
    ...


class FlvTagError(FlvDataError):
    ...


class FlvStreamCorruptedError(Exception):
    ...


class FlvFileCorruptedError(Exception):
    ...



import io
import logging
from abc import ABC, abstractmethod
from datetime import datetime
from threading import Thread
from typing import Any, Iterator, List, Optional, Tuple, Union

import requests
import urllib3
from reactivex import abc
from reactivex.typing import StartableFactory, StartableTarget

from blrec.bili.live import Live
from blrec.bili.typing import QualityNumber, StreamFormat
from blrec.event.event_emitter import EventEmitter, EventListener
from blrec.flv import operators as flv_ops
from blrec.flv.operators import StreamProfile
from blrec.flv.utils import format_timestamp
from blrec.logging.room_id import aio_task_with_room_id
from blrec.setting.typing import RecordingMode
from blrec.utils.mixins import AsyncCooperationMixin, AsyncStoppableMixin

from . import operators as core_ops
from .metadata_provider import MetadataProvider
from .path_provider import PathProvider
from .stream_param_holder import StreamParamHolder

__all__ = ('StreamRecorderImpl',)


logger = logging.getLogger(__name__)
logging.getLogger(urllib3.__name__).setLevel(logging.WARNING)


class StreamRecorderEventListener(EventListener):
    async def on_video_file_created(self, path: str, record_start_time: int) -> None:
        ...

    async def on_video_file_completed(self, path: str) -> None:
        ...

    async def on_stream_recording_interrupted(self, duratin: float) -> None:
        ...

    async def on_stream_recording_recovered(self, timestamp: int) -> None:
        ...

    async def on_stream_recording_completed(self) -> None:
        ...


class StreamRecorderImpl(
    EventEmitter[StreamRecorderEventListener],
    AsyncCooperationMixin,
    AsyncStoppableMixin,
    ABC,
):
    def __init__(
        self,
        live: Live,
        out_dir: str,
        path_template: str,
        *,
        stream_format: StreamFormat = 'flv',
        recording_mode: RecordingMode = 'standard',
        quality_number: QualityNumber = 10000,
        buffer_size: Optional[int] = None,
        read_timeout: Optional[int] = None,
        disconnection_timeout: Optional[int] = None,
        filesize_limit: int = 0,
        duration_limit: int = 0,
    ) -> None:
        super().__init__()

        self._live = live
        self._session = requests.Session()

        self._recording_mode = recording_mode
        self._buffer_size = buffer_size or io.DEFAULT_BUFFER_SIZE
        self._read_timeout = read_timeout or 3
        self._filesize_limit = filesize_limit
        self._duration_limit = duration_limit

        self._stream_param_holder = StreamParamHolder(
            stream_format=stream_format, quality_number=quality_number
        )
        self._stream_url_resolver = core_ops.StreamURLResolver(
            live, self._stream_param_holder
        )
        self._progress_bar = core_ops.ProgressBar(live)
        self._metadata_provider = MetadataProvider(live, self)
        self._path_provider = PathProvider(live, out_dir, path_template)
        self._rec_statistics = core_ops.SizedStatistics()
        self._dl_statistics: Union[core_ops.StreamStatistics, core_ops.SizedStatistics]

        self._request_exception_handler = core_ops.RequestExceptionHandler()
        self._connection_error_handler = core_ops.ConnectionErrorHandler(
            live, disconnection_timeout=disconnection_timeout
        )
        self._exception_handler = core_ops.ExceptionHandler()

        self._subscription: abc.DisposableBase
        self._completed: bool = False

        self._threads: List[Thread] = []
        self._files: List[str] = []
        self._stream_profile: StreamProfile = {}
        self._record_start_time: Optional[int] = None
        self._stream_available_time: Optional[int] = None
        self._hls_stream_available_time: Optional[int] = None

    @property
    def stream_url(self) -> str:
        return self._stream_url_resolver.stream_url

    @property
    def stream_host(self) -> str:
        return self._stream_url_resolver.stream_host

    @property
    def record_start_time(self) -> Optional[int]:
        return self._record_start_time

    @property
    def stream_available_time(self) -> Optional[int]:
        return self._stream_available_time

    @stream_available_time.setter
    def stream_available_time(self, ts: Optional[int]) -> None:
        self._stream_available_time = ts

    @property
    def hls_stream_available_time(self) -> Optional[int]:
        return self._hls_stream_available_time

    @hls_stream_available_time.setter
    def hls_stream_available_time(self, ts: Optional[int]) -> None:
        self._hls_stream_available_time = ts

    @property
    def dl_total(self) -> int:
        return self._dl_statistics.count

    @property
    def dl_rate(self) -> float:
        return self._dl_statistics.rate

    @property
    def rec_elapsed(self) -> float:
        return self._rec_statistics.elapsed

    @property
    def rec_total(self) -> int:
        return self._rec_statistics.count

    @property
    def rec_rate(self) -> float:
        return self._rec_statistics.rate

    @property
    def out_dir(self) -> str:
        return self._path_provider.out_dir

    @out_dir.setter
    def out_dir(self, value: str) -> None:
        self._path_provider.out_dir = value

    @property
    def path_template(self) -> str:
        return self._path_provider.path_template

    @path_template.setter
    def path_template(self, value: str) -> None:
        self._path_provider.path_template = value

    @property
    def stream_format(self) -> StreamFormat:
        return self._stream_param_holder.stream_format

    @property
    def recording_mode(self) -> RecordingMode:
        return self._recording_mode

    @property
    def quality_number(self) -> QualityNumber:
        return self._stream_param_holder.quality_number

    @quality_number.setter
    def quality_number(self, value: QualityNumber) -> None:
        self._stream_param_holder.quality_number = value

    @property
    def real_quality_number(self) -> Optional[QualityNumber]:
        if self.stopped:
            return None
        return self._stream_param_holder.real_quality_number

    @property
    def filesize_limit(self) -> int:
        return self._filesize_limit

    @filesize_limit.setter
    def filesize_limit(self, value: int) -> None:
        self._filesize_limit = value

    @property
    def duration_limit(self) -> int:
        return self._duration_limit

    @duration_limit.setter
    def duration_limit(self, value: int) -> None:
        self._duration_limit = value

    @property
    def read_timeout(self) -> int:
        return self._read_timeout

    @read_timeout.setter
    def read_timeout(self, value: int) -> None:
        self._read_timeout = value

    @property
    def disconnection_timeout(self) -> int:
        return self._connection_error_handler.disconnection_timeout

    @disconnection_timeout.setter
    def disconnection_timeout(self, value: int) -> None:
        self._connection_error_handler.disconnection_timeout = value

    @property
    def buffer_size(self) -> int:
        return self._buffer_size

    @buffer_size.setter
    def buffer_size(self, value: int) -> None:
        self._buffer_size = value

    @property
    def recording_path(self) -> Optional[str]:
        return ''

    @property
    def metadata(self) -> Optional[flv_ops.MetaData]:
        return None

    @property
    def stream_profile(self) -> StreamProfile:
        return self._stream_profile

    def has_file(self) -> bool:
        return bool(self._files)

    def get_files(self) -> Iterator[str]:
        yield from iter(self._files)

    def clear_files(self) -> None:
        self._files.clear()

    def can_cut_stream(self) -> bool:
        return False

    def cut_stream(self) -> bool:
        return False

    def update_progress_bar_info(self) -> None:
        self._progress_bar.update_bar_info()

    def _reset(self) -> None:
        self._files.clear()
        self._stream_profile = {}
        self._record_start_time = None
        self._completed = False

    async def _do_start(self) -> None:
        logger.debug('Starting stream recorder...')
        self._on_start()
        self._reset()
        self._run()
        logger.debug('Started stream recorder')

    async def _do_stop(self) -> None:
        logger.debug('Stopping stream recorder...')
        self._stream_param_holder.cancel()
        thread = self._thread_factory('StreamRecorderDisposer')(self._dispose)
        thread.start()
        for thread in self._threads:
            await self._loop.run_in_executor(None, thread.join, 30)
        self._threads.clear()
        self._on_stop()
        logger.debug('Stopped stream recorder')

    def _on_start(self) -> None:
        pass

    def _on_stop(self) -> None:
        pass

    @abstractmethod
    def _run(self) -> None:
        raise NotImplementedError()

    def _thread_factory(self, name: str) -> StartableFactory:
        def factory(target: StartableTarget) -> Thread:
            thread = Thread(
                target=target, daemon=True, name=f'{name}::{self._live.room_id}'
            )
            self._threads.append(thread)
            return thread

        return factory

    def _dispose(self) -> None:
        self._subscription.dispose()
        del self._subscription
        self._on_completed()

    def _on_completed(self) -> None:
        if self._completed:
            return
        self._completed = True

        self._dl_statistics.freeze()
        self._rec_statistics.freeze()
        self._emit_event('stream_recording_completed')

    def _on_profile_updated(self, profile: StreamProfile) -> None:
        logger.debug(f'Stream profile: {profile}')
        self._stream_profile = profile

    def _on_video_file_opened(self, args: Tuple[str, int]) -> None:
        logger.info(f"Video file created: '{args[0]}'")
        self._files.append(args[0])
        self._record_start_time = args[1]
        self._emit_event('video_file_created', *args)

    def _on_video_file_closed(self, path: str) -> None:
        logger.info(f"Video file completed: '{path}'")
        self._emit_event('video_file_completed', path)

    def _on_recording_interrupted(self, duration: float) -> None:
        duration_string = format_timestamp(int(duration * 1000))
        logger.info(f'Recording interrupted, current duration: {duration_string}')
        self._emit_event('stream_recording_interrupted', duration)

    def _on_recording_recovered(self, timestamp: int) -> None:
        datetime_string = datetime.fromtimestamp(timestamp).isoformat()
        logger.info(f'Recording recovered, current date time {(datetime_string)}')
        self._emit_event('stream_recording_recovered', timestamp)

    def _emit_event(self, name: str, *args: Any, **kwds: Any) -> None:
        self._run_coroutine(self._emit(name, *args, **kwds))

    @aio_task_with_room_id
    async def _emit(self, *args: Any, **kwds: Any) -> None:  # type: ignore
        await super()._emit(*args, **kwds)



from __future__ import annotations

from collections import OrderedDict
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Any, Dict, Union

from blrec import __github__, __prog__, __version__
from blrec.bili.helpers import get_quality_name
from blrec.bili.live import Live

if TYPE_CHECKING:
    from .stream_recorder_impl import StreamRecorderImpl

__all__ = ('MetadataProvider',)


class MetadataProvider:
    def __init__(self, live: Live, stream_recorder: StreamRecorderImpl) -> None:
        super().__init__()
        self._live = live
        self._stream_recorder = stream_recorder

    def __call__(self, original_metadata: Dict[str, Any]) -> Dict[str, Any]:
        return self._make_metadata(original_metadata)

    def _make_metadata(self, original_metadata: Dict[str, Any]) -> Dict[str, Any]:
        tz = timezone(timedelta(hours=8))
        live_start_time = datetime.fromtimestamp(
            self._live.room_info.live_start_time, tz
        )
        if self._stream_recorder.stream_available_time is None:
            stream_available_time: Union[datetime, str] = 'N/A'
        else:
            stream_available_time = datetime.fromtimestamp(
                self._stream_recorder.stream_available_time, tz
            )
        if self._stream_recorder.hls_stream_available_time is None:
            hls_stream_available_time: Union[datetime, str] = 'N/A'
        else:
            hls_stream_available_time = datetime.fromtimestamp(
                self._stream_recorder.hls_stream_available_time, tz
            )
        if self._stream_recorder.record_start_time is None:
            record_start_time: Union[datetime, str] = 'N/A'
        else:
            record_start_time = datetime.fromtimestamp(
                self._stream_recorder.record_start_time, tz
            )

        assert self._stream_recorder.real_quality_number is not None
        stream_quality = '{} ({}{})'.format(
            get_quality_name(self._stream_recorder.real_quality_number),
            self._stream_recorder.real_quality_number,
            ', bluray' if '_bluray' in self._stream_recorder.stream_url else '',
        )

        if self._stream_recorder.recording_mode == 'standard':
            recording_mode_desc = '标准'
        elif self._stream_recorder.recording_mode == 'raw':
            recording_mode_desc = '原始'
        else:
            recording_mode_desc = '？？'

        return {
            'Title': self._live.room_info.title,
            'Artist': self._live.user_info.name,
            'Date': str(live_start_time),
            'Comment': f'''\
B站直播录像
主播：{self._live.user_info.name}
标题：{self._live.room_info.title}
分区：{self._live.room_info.parent_area_name} - {self._live.room_info.area_name}
房间号：{self._live.room_info.room_id}
开播时间：{live_start_time}
开始推流时间: {stream_available_time}
HLS流可用时间: {hls_stream_available_time}
开始录制时间: {record_start_time}
流主机: {self._stream_recorder.stream_host}
流格式：{self._stream_recorder.stream_format}
流画质：{stream_quality}
录制模式: {recording_mode_desc}
录制程序：{__prog__} v{__version__} {__github__}''',
            'description': OrderedDict(
                {
                    'UserId': str(self._live.user_info.uid),
                    'UserName': self._live.user_info.name,
                    'RoomId': str(self._live.room_info.room_id),
                    'RoomTitle': self._live.room_info.title,
                    'Area': self._live.room_info.area_name,
                    'ParentArea': self._live.room_info.parent_area_name,
                    'LiveStartTime': str(live_start_time),
                    'StreamAvailableTime': str(stream_available_time),
                    'HLSStreamAvailableTime': str(hls_stream_available_time),
                    'RecordStartTime': str(record_start_time),
                    'StreamHost': self._stream_recorder.stream_host,
                    'StreamFormat': self._stream_recorder.stream_format,
                    'StreamQuality': stream_quality,
                    'RecordingMode': self._stream_recorder.recording_mode,
                    'Recorder': f'{__prog__} v{__version__} {__github__}',
                }
            ),
        }



import logging
from io import BytesIO
from typing import Any, BinaryIO, Mapping, TypedDict

from .amf import AMFReader, AMFWriter

__all__ = (
    'load',
    'loads',
    'dump',
    'dumps',
    'ScriptData',
    'ScriptDataParser',
    'ScriptDataDumper',
)

logger = logging.getLogger(__name__)


class ScriptData(TypedDict):
    name: str
    value: Any


class ScriptDataParser:
    def __init__(self, stream: BinaryIO) -> None:
        self._reader = AMFReader(stream)

    def parse(self) -> ScriptData:
        name = self._parse_name()
        try:
            value = self._parse_value()
        except EOFError:
            logger.debug(f'No script data: {name}')
            value = {}
        if not isinstance(value, dict):
            if name == 'onMetaData':
                logger.debug(f'Invalid onMetaData: {value}')
                value = {}
            else:
                logger.debug(f'Unusual script data: {name}, {value}')
        return ScriptData(name=name, value=value)

    def _parse_name(self) -> str:
        value = self._reader.read_value()
        assert isinstance(value, str)
        return value

    def _parse_value(self) -> Any:
        return self._reader.read_value()


class ScriptDataDumper:
    def __init__(self, stream: BinaryIO) -> None:
        self._writer = AMFWriter(stream)

    def dump(self, script_data: ScriptData) -> None:
        self._dump_name(script_data['name'])
        self._dump_value(script_data['value'])

    def _dump_name(self, name: str) -> None:
        self._writer.write_value(name)

    def _dump_value(self, value: Mapping[str, Any]) -> None:
        self._writer.write_value(value)


def load(data: bytes) -> ScriptData:
    return loads(BytesIO(data))


def loads(stream: BinaryIO) -> ScriptData:
    return ScriptDataParser(stream).parse()


def dump(script_data: ScriptData) -> bytes:
    stream = BytesIO()
    dumps(script_data, stream)
    return stream.getvalue()


def dumps(script_data: ScriptData, stream: BinaryIO) -> None:
    ScriptDataDumper(stream).dump(script_data)



import json
import logging
from contextlib import suppress
from typing import List, Optional

import attr

from ..flv import operators as flv_ops
from ..path import extra_metadata_path
from ..utils.mixins import SwitchableMixin

__all__ = ('MetadataDumper',)

logger = logging.getLogger(__name__)


class MetadataDumper(SwitchableMixin):
    def __init__(
        self,
        dumper: flv_ops.Dumper,
        analyser: flv_ops.Analyser,
        joinpoint_extractor: flv_ops.JoinPointExtractor,
    ) -> None:
        super().__init__()
        self._dumper = dumper
        self._analyser = analyser
        self._joinpoint_extractor = joinpoint_extractor
        self._reset()

    def _reset(self) -> None:
        self._last_metadata: Optional[flv_ops.MetaData] = None
        self._last_join_points: Optional[List[flv_ops.JoinPoint]] = None

    def _do_enable(self) -> None:
        self._metadata_subscription = self._analyser.metadatas.subscribe(
            on_next=self._update_metadata
        )
        self._join_points_subscription = (
            self._joinpoint_extractor.join_points.subscribe(self._update_join_points)
        )
        self._file_closed_subscription = self._dumper.file_closed.subscribe(
            self._dump_metadata
        )
        self._reset()
        logger.debug('Enabled metadata dumper')

    def _do_disable(self) -> None:
        with suppress(Exception):
            self._metadata_subscription.dispose()
        with suppress(Exception):
            self._join_points_subscription.dispose()
        with suppress(Exception):
            self._file_closed_subscription.dispose()
        self._reset()
        logger.debug('Disabled metadata dumper')

    def _update_metadata(self, metadata: Optional[flv_ops.MetaData]) -> None:
        self._last_metadata = metadata

    def _update_join_points(self, join_points: List[flv_ops.JoinPoint]) -> None:
        self._last_join_points = join_points

    def _dump_metadata(self, video_path: str) -> None:
        path = extra_metadata_path(video_path)
        logger.debug(f"Dumping metadata to file: '{path}'")

        if self._last_metadata is not None:
            data = attr.asdict(self._last_metadata, filter=lambda a, v: v is not None)
        else:
            data = {}
            logger.warning('The metadata may be lost duo to something went wrong')

        if self._last_join_points is not None:
            data['joinpoints'] = list(
                map(lambda p: p.to_metadata_value(), self._last_join_points)
            )
        else:
            data['joinpoints'] = []
            logger.warning('The joinpoints may be lost duo to something went wrong')

        try:
            with open(path, 'wt', encoding='utf8') as file:
                json.dump(data, file)
        except Exception as e:
            logger.error(f'Failed to dump metadata: {e}')
        else:
            logger.debug(f"Successfully dumped metadata to file: '{path}'")



import asyncio
import logging

from fastapi import APIRouter

from ... import __prog__
from ...application import Application
from ...update.helpers import get_latest_version_string


logger = logging.getLogger(__name__)


app: Application = None  # type: ignore  # bypass flake8 F821

router = APIRouter(
    prefix='/api/v1/update',
    tags=['update'],
)


@router.get(
    '/version/latest',
)
async def get_latest_version() -> str:
    try:
        return await get_latest_version_string(__prog__) or ''
    except asyncio.TimeoutError:
        logger.warning(
            'Timeout error occurred while getting the latest version string'
        )
        return ''




import aiohttp

from .api import WebApi
from .typing import ResponseData, QualityNumber
from .exceptions import ApiRequestError
from ..exception import NotFoundError


__all__ = 'room_init', 'ensure_room_id'


async def room_init(room_id: int) -> ResponseData:
    async with aiohttp.ClientSession(raise_for_status=True) as session:
        api = WebApi(session)
        return await api.room_init(room_id)


async def ensure_room_id(room_id: int) -> int:
    """Ensure room id is valid and is the real room id"""
    try:
        result = await room_init(room_id)
    except ApiRequestError as e:
        if e.code == 60004:
            raise NotFoundError(f'the room {room_id} not existed')
        else:
            raise
    else:
        return result['room_id']


def get_quality_name(qn: QualityNumber) -> str:
    QUALITY_MAPPING = {
        20000: '4K',
        10000: '原画',
        401: '蓝光(杜比)',
        400: '蓝光',
        250: '超清',
        150: '高清',
        80: '流畅',
    }
    return QUALITY_MAPPING.get(qn, '')



from __future__ import annotations

import uuid
from abc import ABC
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Any, Dict, Generic, List, Type, TypeVar

import attr

if TYPE_CHECKING:
    from ..bili.models import UserInfo, RoomInfo
    from ..disk_space.space_monitor import DiskUsage

from ..exception import format_exception

_D = TypeVar('_D', bound='BaseEventData')
_E = TypeVar('_E')
# XXX error: Invalid self argument ...
# _E = TypeVar('_E', bound='BaseEvent[BaseEventData]')


class BaseEventData(ABC):
    pass


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class BaseEvent(Generic[_D]):
    type: str
    id: uuid.UUID
    date: datetime
    data: _D

    @classmethod
    def from_data(cls: Type[_E], data: _D) -> _E:
        return cls(  # type: ignore
            id=uuid.uuid1(), date=datetime.now(timezone(timedelta(hours=8))), data=data
        )

    def asdict(self) -> Dict[str, Any]:
        return attr.asdict(self, value_serializer=value_serializer)


def value_serializer(instance, attribute, value):  # type: ignore
    if isinstance(value, (uuid.UUID, datetime)):
        return str(value)
    return value


@attr.s(auto_attribs=True, slots=True, frozen=True)
class LiveBeganEventData(BaseEventData):
    user_info: UserInfo
    room_info: RoomInfo


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class LiveBeganEvent(BaseEvent[LiveBeganEventData]):
    type: str = 'LiveBeganEvent'
    data: LiveBeganEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class LiveEndedEventData(LiveBeganEventData):
    pass


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class LiveEndedEvent(BaseEvent[LiveEndedEventData]):
    type: str = 'LiveEndedEvent'
    data: LiveEndedEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class RoomChangeEventData(BaseEventData):
    room_info: RoomInfo


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class RoomChangeEvent(BaseEvent[RoomChangeEventData]):
    type: str = 'RoomChangeEvent'
    data: RoomChangeEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class RecordingStartedEventData(BaseEventData):
    room_info: RoomInfo


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class RecordingStartedEvent(BaseEvent[RecordingStartedEventData]):
    type: str = 'RecordingStartedEvent'
    data: RecordingStartedEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class RecordingFinishedEventData(BaseEventData):
    room_info: RoomInfo


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class RecordingFinishedEvent(BaseEvent[RecordingFinishedEventData]):
    type: str = 'RecordingFinishedEvent'
    data: RecordingFinishedEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class RecordingCancelledEventData(BaseEventData):
    room_info: RoomInfo


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class RecordingCancelledEvent(BaseEvent[RecordingCancelledEventData]):
    type: str = 'RecordingCancelledEvent'
    data: RecordingCancelledEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class VideoFileCreatedEventData(BaseEventData):
    room_id: int
    path: str


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class VideoFileCreatedEvent(BaseEvent[VideoFileCreatedEventData]):
    type: str = 'VideoFileCreatedEvent'
    data: VideoFileCreatedEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class VideoFileCompletedEventData(BaseEventData):
    room_id: int
    path: str


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class VideoFileCompletedEvent(BaseEvent[VideoFileCompletedEventData]):
    type: str = 'VideoFileCompletedEvent'
    data: VideoFileCompletedEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class DanmakuFileCreatedEventData(BaseEventData):
    room_id: int
    path: str


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class DanmakuFileCreatedEvent(BaseEvent[DanmakuFileCreatedEventData]):
    type: str = 'DanmakuFileCreatedEvent'
    data: DanmakuFileCreatedEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class DanmakuFileCompletedEventData(BaseEventData):
    room_id: int
    path: str


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class DanmakuFileCompletedEvent(BaseEvent[DanmakuFileCompletedEventData]):
    type: str = 'DanmakuFileCompletedEvent'
    data: DanmakuFileCompletedEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class RawDanmakuFileCreatedEventData(BaseEventData):
    room_id: int
    path: str


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class RawDanmakuFileCreatedEvent(BaseEvent[RawDanmakuFileCreatedEventData]):
    type: str = 'RawDanmakuFileCreatedEvent'
    data: RawDanmakuFileCreatedEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class RawDanmakuFileCompletedEventData(BaseEventData):
    room_id: int
    path: str


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class RawDanmakuFileCompletedEvent(BaseEvent[RawDanmakuFileCompletedEventData]):
    type: str = 'RawDanmakuFileCompletedEvent'
    data: RawDanmakuFileCompletedEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class CoverImageDownloadedEventData(BaseEventData):
    room_id: int
    path: str


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class CoverImageDownloadedEvent(BaseEvent[CoverImageDownloadedEventData]):
    type: str = 'CoverImageDownloadedEvent'
    data: CoverImageDownloadedEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class VideoPostprocessingCompletedEventData(BaseEventData):
    room_id: int
    path: str


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class VideoPostprocessingCompletedEvent(
    BaseEvent[VideoPostprocessingCompletedEventData]
):
    type: str = 'VideoPostprocessingCompletedEvent'
    data: VideoPostprocessingCompletedEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class PostprocessingCompletedEventData(BaseEventData):
    room_id: int
    files: List[str]


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class PostprocessingCompletedEvent(BaseEvent[PostprocessingCompletedEventData]):
    type: str = 'PostprocessingCompletedEvent'
    data: PostprocessingCompletedEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class SpaceNoEnoughEventData(BaseEventData):
    path: str
    threshold: int
    usage: DiskUsage


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class SpaceNoEnoughEvent(BaseEvent[SpaceNoEnoughEventData]):
    type: str = 'SpaceNoEnoughEvent'
    data: SpaceNoEnoughEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class ErrorData(BaseEventData):
    name: str
    detail: str

    @classmethod
    def from_exc(cls, exc: BaseException) -> ErrorData:
        return cls(name=type(exc).__name__, detail=format_exception(exc))


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class Error(BaseEvent[ErrorData]):
    type: str = 'Error'
    data: ErrorData



from typing import Protocol, runtime_checkable


__all__ = (
    'ReadableStream',
    'AsyncReadableStream',
    'WritableStream',
    'AsyncWritableStream',
    'RandomIO',
    'AsyncRandomIO',
)


@runtime_checkable
class ReadableStream(Protocol):
    def read(self, size: int = -1) -> bytes:
        ...


@runtime_checkable
class AsyncReadableStream(Protocol):
    async def read(self, size: int = -1) -> bytes:
        ...


@runtime_checkable
class WritableStream(Protocol):
    def write(self, data: bytes) -> int:
        ...


@runtime_checkable
class AsyncWritableStream(Protocol):
    async def write(self, data: bytes) -> int:
        ...


@runtime_checkable
class RandomIO(ReadableStream, WritableStream, Protocol):
    def tell(self) -> int:
        ...

    def seek(self, offset: int, whence: int = 0) -> int:
        ...

    def truncate(self, size: int = None) -> int:
        ...


@runtime_checkable
class AsyncRandomIO(AsyncReadableStream, AsyncWritableStream, Protocol):
    async def tell(self) -> int:
        ...

    async def seek(self, offset: int, whence: int = 0) -> int:
        ...

    async def truncate(self, size: int = None) -> int:
        ...



import re
from enum import IntEnum
from datetime import datetime
from typing import cast

import attr
from lxml import html
from lxml.html.clean import clean_html

from .typing import ResponseData
from ..utils.url import ensure_scheme


__all__ = 'LiveStatus', 'RoomInfo', 'UserInfo'


class LiveStatus(IntEnum):
    PREPARING = 0
    LIVE = 1
    ROUND = 2


@attr.s(auto_attribs=True, frozen=True, slots=True)
class RoomInfo:
    uid: int
    room_id: int
    short_room_id: int
    area_id: int
    area_name: str
    parent_area_id: int
    parent_area_name: str
    live_status: LiveStatus
    live_start_time: int  # a integer in seconds
    online: int
    title: str
    cover: str
    tags: str
    description: str

    @staticmethod
    def from_data(data: ResponseData) -> 'RoomInfo':
        if (timestamp := data.get('live_start_time')) is not None:
            live_start_time = cast(int, timestamp)
        elif (time_string := data.get('live_time')) is not None:
            if time_string == '0000-00-00 00:00:00':
                live_start_time = 0
            else:
                dt = datetime.fromisoformat(time_string)
                live_start_time = int(dt.timestamp())
        else:
            raise ValueError(f'Failed to init live_start_time: {data}')

        if (cover := data.get('cover') or data.get('user_cover', '')):
            cover = ensure_scheme(cover, 'https')

        if (description := data['description']):
            description = re.sub(r'<br\s*/?>', '\n', description)
            tree = html.fromstring(description)
            description = clean_html(tree).text_content().strip()

        return RoomInfo(
            uid=data['uid'],
            room_id=int(data['room_id']),
            short_room_id=int(data['short_id']),
            area_id=data['area_id'],
            area_name=data['area_name'],
            parent_area_id=data['parent_area_id'],
            parent_area_name=data['parent_area_name'],
            live_status=LiveStatus(data['live_status']),
            live_start_time=live_start_time,
            online=int(data['online']),
            title=data['title'],
            cover=cover,
            tags=data['tags'],
            description=description,
        )


@attr.s(auto_attribs=True, frozen=True, slots=True)
class UserInfo:
    name: str
    gender: str
    face: str
    uid: int
    level: int
    sign: str

    @staticmethod
    def from_web_api_data(data: ResponseData) -> 'UserInfo':
        return UserInfo(
            name=data['name'],
            gender=data['sex'],
            face=ensure_scheme(data['face'], 'https'),
            uid=data['mid'],
            level=data['level'],
            sign=data['sign'],
        )

    @staticmethod
    def from_app_api_data(data: ResponseData) -> 'UserInfo':
        card = data['card']
        return UserInfo(
            name=card['name'],
            gender=card['sex'],
            face=ensure_scheme(card['face'], 'https'),
            uid=card['mid'],
            level=card['level_info']['current_level'],
            sign=card['sign'],
        )
